/*
 * Date: 2003-01-16
 *
 */

package tein.pt2.intro;

//Imports from Java
import java.util.HashMap;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.StringTokenizer;
import java.io.*;

/**
 * This class will demonstrate the use of the following java constructs:
 * - package. The use of packages to structurize java code.
 * - import. The use of imports to access classes in other packages.
 * - JavaDoc. The use of javadoc tags.
 *
 * The class implements a simple phone book. A dataentry consists of a name
 * and a phone number. The operations on this phone book is add, delete, search
 * and list.
 * All entries are saved on file to make it persistent. A data entry is
 * stored in the file as a single row containing a name and a phone number.
 * To make it more complex the phone book supports multiple entries with
 * the same name.
 */
public class MWPhoneBook {


    // Constants
    public static final byte[] NEW_LINE = "\n".getBytes();

    // Class Variables
    // Instance Variables

    /** The Hash map used for looking up phone numbers using names. */
    private HashMap iNameMap;

    /**
     * This is the constructor used to create the phone book object.
     */
    public MWPhoneBook() {
        iNameMap = new HashMap();

    }

    /**
     * This method will load the content of a phonebook file
     * into the program. It is possible to load several files.
     * @param fileName
     *        The name of the file to load.
     * @param purgeOldData
     *        A boolean telling whether to purge all current data
     *        loaded into the program or not.
     */
    public void load(String fileName, boolean purgeOldData) {
        BufferedReader reader;
        String line = null;

        // If we are to discard old data.
        if( purgeOldData) {
            iNameMap.clear();
        }

        //open file for reading.

        try {
            reader = new BufferedReader( new FileReader( fileName));

            // Read an entry from the file.
            do {
                line = reader.readLine();

                //Check that we havent reached EOF
                if( line != null) {
                    add(MWPhoneBookEntry.fromString( line));
                }
            } while( line != null);
        }
        catch(IOException e) {
            System.out.println("MWPhoneBook.load/2: Could not open file for reading: ["+fileName+"]");
        }
    }

    /**
     * Saves all entries to file, sorted.
     * @param fileName
     *        The name of the file to store all entries in.
     */
    public void save(String fileName) {
        BufferedOutputStream out;

        // Put all entries into an array and sort it.
        // This array could be created at load time since
        // a sorted array is also used when listing all entries.
        // This is left to the reader to modify.
        MWPhoneBookEntry[] entries = toArray();

        try {
            out = new BufferedOutputStream(new FileOutputStream(fileName));

            for(int i=0; i< entries.length; i++) {
                out.write(entries[i].toString().getBytes());
                out.write(NEW_LINE);
            }
            out.close();
        }

        catch(IOException e) {
            System.out.println("MWPhoneBook.save/1: Could not open file for writing: ["+fileName+"]");
        }
    }

    /**
     * This method will return a list of all entries as a string.
     * @return A list of all entries as a single string.
     */
    public String toString() {
        //Again a sorted array is needed. See note in save method.
        MWPhoneBookEntry[] entries = toArray();
        return toString( entries);
    }

    public String toString(MWPhoneBookEntry[] entries) {
        StringBuffer buffer = new StringBuffer(1024);

        for(int i=0; i< entries.length; i++) {
            buffer.append(entries[i].toString()).append("\n");
        }
        return buffer.toString();
    }

    /**
     * A helper method returning all entries in a soreted array.
     * @return A sorted array containing all entries.
     */
    private MWPhoneBookEntry[] toArray() {
        Object[] entryLists = iNameMap.values().toArray();
        ArrayList bigList = new ArrayList();
        MWPhoneBookEntry[] answer = null;

        for(int i=0; i< entryLists.length; i++) {
            bigList.addAll((ArrayList) entryLists[i]);
        }

        answer = new MWPhoneBookEntry[ bigList.size()];
        bigList.toArray( answer);
        Arrays.sort( answer);
        return answer;
    }

    /**
     * Searches for entries with the given key. For now the key
     * is the name of a person
     * @param key
     *        The key to look for among the entries.
     * @return All entries with the given key.
     */
    public MWPhoneBookEntry[] search(String key) {
        ArrayList entryList = (ArrayList) iNameMap.get(key.toLowerCase());
        MWPhoneBookEntry[] answer = null;

        if( entryList != null) {
            answer = new MWPhoneBookEntry[ entryList.size()];
            entryList.toArray(answer);
        }
        return answer;
    }

    /**
     * Adds an entry to the phone book.
     * @param entry
     *        The entry to add.
     */
    public void add( MWPhoneBookEntry entry) {

        ArrayList entryList = (ArrayList) iNameMap.get( entry.getName().toLowerCase());

        if( entryList == null) {
            entryList = new ArrayList();
            iNameMap.put( entry.getName().toLowerCase(), entryList);
        }
        entryList.add( entry);
    }

    /**
     * Removes all entries with the specified key/name.
     * @param key
     *        The key to the entries to be deleted.
     */
    public void remove( String key) {
        iNameMap.remove(key.toLowerCase());
    }

    /**
     * This method interprets user input and performs the actions.
     * @param commandLine
     *        The user input to interpret.
     * @return true iff the interpreter expects the program to go on.
     *        false iff the user issued the exit command.
     */
    public boolean interpret(String commandLine) {
        StringTokenizer tokenizer = new StringTokenizer(commandLine);
        String token = null;
        boolean answer = true;

        while(tokenizer.hasMoreTokens() && answer) {
            token = tokenizer.nextToken();

            if( token.equalsIgnoreCase("+") && tokenizer.countTokens() >= 2) {
                add(new MWPhoneBookEntry(tokenizer.nextToken(), tokenizer.nextToken()));
                System.out.println("");
            }
            else if(token.equalsIgnoreCase("-") && tokenizer.countTokens() >= 1) {
                remove(tokenizer.nextToken());
                System.out.println("");
            }
            else if(token.equalsIgnoreCase("?") && tokenizer.countTokens() >= 1) {
                MWPhoneBookEntry[] entries = search(tokenizer.nextToken());

                if(entries != null) {
                    System.out.println("--- Entries Found ---");
                    System.out.println(toString(entries));
                }
                else {
                    System.out.println("No entries found!");
                }
            }
            else if( token.equalsIgnoreCase("save") && tokenizer.countTokens() >= 1) {
                save(tokenizer.nextToken());
                System.out.println("");
            }
            else if( token.equalsIgnoreCase("load") && tokenizer.countTokens() >= 2) {
                load(tokenizer.nextToken(), Boolean.valueOf(tokenizer.nextToken()).booleanValue());
                System.out.println("");
            }
            else if( token.equalsIgnoreCase("exit")) {
                answer = false;
                System.out.println("");
            }
            else if( token.equalsIgnoreCase("list")) {
                System.out.println("--- All Entries in Phone Book ---");
                System.out.println( toString());
            }
            else if( token.equalsIgnoreCase("help")) {
                System.out.println(" + <name> <phone number> : Adds a new entry to the phone book.");
                System.out.println(" - <name>                : Removes all entries with the specified name.");
                System.out.println(" ? <name>                : Searches for all entries with the specified name.");
                System.out.println(" list                    : Lists all entries in the phone book.");
                System.out.println(" save <fileName>         : Saves all entries in a file");
                System.out.println(" load <fileName> <purge> : Loads entries from file. purge is true|false specifying if current entries should be purged.");
                System.out.println(" exit                    : Exits the program, without saving data.");
                System.out.println(" help                    : Prints this text.");
                System.out.println(" At startup the program accepts a file name as an argument to load entries from.");
            }
            else {
                System.out.println("Unknown command: ["+token+"]. Write \"help\" for additional info about valid commands");
            }
        }
        return answer;
    }

    public static void main( String[] argv) {

        try {
            BufferedReader reader = new BufferedReader(new InputStreamReader(System.in));
            MWPhoneBook book = new MWPhoneBook();

            if( argv.length >= 1) {
                book.load( argv[0], true);
            }

            do {
                System.out.print("book:\\>");
            }
            while( book.interpret( reader.readLine()));

        }
        catch( IOException e) {
            System.out.println("MWPhoneBook.main/1: Exception while reading from command line");
            e.printStackTrace();
        }
    }
}
